#!/usr/bin/perl

#
# dxftoobj -- convert autocad .dxf files to wavefront's .obj format
#
#   options:
#      --verbose          produce more information about conversion
#      --disable-text     disable conversion of TEXT entities
#      --bytype           groups entities by their type
#      --arc-segments=N   resolution of ARC (and CIRCLE) entities
#      --postscript       output PostScript
#      --obj              output OBJ (default)
#      --mdl             output MDL (for I3DM Performer modeller)
#
# 06/94 chas@ohm
#   support for SOLID, LINE, 3DFACE, ARC (XY plane only), and POLYLINE
#   support for POLYLINE polygon mesh
#   CIRCLE support via ARC
# 07/94 chas@ohm
#   support for POLYLINE polyface mesh
#   ignore POINT entities (for now)
#   added grouping by LAYER name (in some places)
# 09/95 chas@ohm
#   improved support for POLYLINE polygon mesh
#   fixed handling for "degenerate" 3DFACE entities
#   BLOCK/INSERT support
#   initial support for TEXT entities (via Wavefront's "bit" utility)
#   minor rewrite for modularity
#   added beginnings for postscript output option
# 10/95 chas@ohm
#   support thickness for some objects (LINE and some POLYLINEs)
#   support TRACE objects and fixed SOLID support
#   partial support for defunct group code 38 (elevation)
# 08/96 chas@ohm
#   major rewrite to support other output formats
#   fixed postscript output
#   added mdl output (for I3DM a modeler provided with SGI Performer)
#   a couple change so it runs under perl4 and perl5
#   
# BUGS:
#    Parts of polyline still don't use the driver_ routines
#    Groups 210, 220, 230 are not supported (i.e. rotations)
#    Arcs/circles must lie in the 0,0,1 plane (see above)
#    Thickness, extrusions, and elevations not supported for all entities.
#    Autocad polygons are doublesided.  Some formats don't support this.
#    The postscript output is viewed only along the Z-axis.
#    BLOCK works, but must never occur in the ENTITIES section and you
#       might need to run dxftoobj twice since BLOCK/INSERT order is arbitrary.
#
# TODO:
#    Support 210, 220, 230
#    Postscript output should be AIFM compatiable (see gnuplot)
#

# program defaults...
$dxftoobj_version = "0.7b";
$verbose=0;
$driver="obj";
$groupbytype = 0;
$disable_text = 0;
$arc_segments = 10;

$oldgroup="";
$seen'elevation = 0;

# counts for various items
$dxf_npolyline = 0;
$dxf_npolyline_gmesh = 0;
$dxf_npolyline_fmesh = 0;
$dxf_npolyline_thick = 0;
$dxf_nline = 0;
$dxf_n3dline = 0;
$dxf_nline_thick = 0;
$dxf_nvertex = 0;
$dxf_npoint = 0;
$dxf_n3dface = 0;
$dxf_ntext = 0;
$dxf_nsolid = 0;
$dxf_ntrace = 0;
$dxf_ncircle = 0;
$dxf_narc = 0;
$dxf_ninsert = 0;
$dxf_nblock = 0;

while ($_ = shift(@ARGV) )
{
	check:
	{
		/^--usage/ && ($usage, last check);
		/^--verbose/ && (++$verbose, last check);
		/^--disable-text/ && ($disable_text=1, last check);
		/^--group-by-type/ && ( $groupbytype=1, last check);
		/^--arc-segments/ && do { ($arc_segments) = /.*=(\d+)/ };
		
		/^--postscript/ && ( $driver="ps", last check);
		/^--mdl/ && ( $driver="mdl", last check);
		/^--obj/ && ( $driver="obj", last check);
		
		/^-/ && do {print STDERR "unknown option \"$_\"\n"; last check};

		$dxf = $_;
	}
}

print STDERR "!!! BETA VERSION (indeterminate behavior possible) !!!\n";

# initialize some variables...
$__line__ = 0;
$lastentity = "none";
undef $variable;
undef $lastlayer;

open(DXF,"<$dxf") || do
{
	$dxf="$dxf.dxf"; # try appending .dxf
	open(DXF,"<$dxf") || do
	{
		print STDERR "dxftoobj: failed to open $dxf\n";
		exit;
	};
};

($output = $dxf) =~ s/\.dxf$//;

while(<DXF>)
{
	# group followed by value...
	$group = $_; ++$__line__;
	$group =~ s/[\012\015]+$//;	# chop doesn't remove ^M

	$value = <DXF>; ++$__line__;
	$value =~ s/[\012\015]+$//;

	#
	# basically, discovering and decoding entities are
	# the major part of decoding a dxf file
	#
	$group == 0 && do
	{
		defined $entity && ( $lastentity = $entity );

		case:
		{
			# have any entities been seen yet?
			defined $lastentity || last case;

			# process the last entity
			$lastentity eq "VERTEX" &&   (&dxf_vertex, last case);
			$lastentity eq "3DFACE" &&   (&dxf_3dface, last case);
			$lastentity eq "LINE" &&     (&dxf_line, last case);
			$lastentity eq "3DLINE" &&   (&dxf_3dline, last case);
			$lastentity eq "POLYLINE" && (&dxf_polyline, last case);
			$lastentity eq "SEQEND" &&   (&dxf_seqend, last case);
			$lastentity eq "ARC" &&      (&dxf_arc, last case);
			$lastentity eq "CIRCLE" &&   (&dxf_circle, last case);
			$lastentity eq "SOLID" &&    (&dxf_solid, last case);
			$lastentity eq "TRACE" &&    (&dxf_trace, last case);
			$lastentity eq "POINT" &&    (&dxf_point, last case);
			$lastentity eq "TEXT" &&     (&dxf_text, last case);
		
			$lastentity eq "BLOCK" &&    (&dxf_block, last case);
			$lastentity eq "ENDBLK" &&   (&dxf_endblk, last case);
			$lastentity eq "INSERT" &&   (&dxf_insert, last case);

			# ignore these entities for now
			$lastentity eq "VPORT" && (last case);
			$lastentity eq "LAYER" && (last case);
			$lastentity eq "STYLE" && (last case);
			$lastentity eq "LTYPE" && (last case);
			$lastentity eq "UCS" && (last case);
			$lastentity eq "VIEW" && (last case);
			$lastentity eq "ATTDEF" && (last case);
		
			# these are special and handled elsewhere...
			$lastentity eq "SECTION" && (last case);
			$lastentity eq "ENDSEC" && (last case);
			$lastentity eq "TABLE" && (last case);
			$lastentity eq "ENDTAB" && (last case);

			$lastentity ne "none" &&
				&info("DXF: unsupported ENTITY \"$lastentity\"");
		};

		# now that we have processed the old entity 
		# reset the variables to the default values
		# (this isn't complete by any means at the moment)
		
		undef @integer;
		undef @float;
		undef $layer;
		undef $thickness;
		undef $elevation;

		$x[0] = 0.0; $x[1] = 0.0; $x[2] = 0.0; $x[3] = 0.0; $x[4] = 0.0;
		$y[0] = 0.0; $y[1] = 0.0; $y[2] = 0.0; $y[3] = 0.0; $y[4] = 0.0;
		$z[0] = 0.0; $z[1] = 0.0; $z[2] = 0.0; $z[3] = 0.0; $z[4] = 0.0;

		$extrude_x = 0;	# default values for extrusions
		$extrude_y = 0;
		$extrude_z = 1;

		$entity = $value; # current entity
		$__entity_line = $__line__;

		$entity eq "SECTION" && do
		{
			$section'in = 1;
			$section'name = "";
		};

		$entity eq "TABLE" && do
		{
			$table'in = 1;
			$table'name = "";
		};

		$entity eq "ENDSEC" && ($section'in = 0);
		$entity eq "ENDTAB" && ($section'in = 0);

		$entity eq "EOF" && ( &dxf_eof, exit 0);

		next;
	};

	$group == 2 && do
	{
		$name = $value;

		if ( $entity eq "SECTION" && $section'name eq "" )
		{
			$section'name = $name;
			&info("dxftoobj: processing SECTION $section'name...");
			if ( $section'name eq "ENTITIES" )
				{ &dxf_init; }
		}

		if ( $entity eq "TABLE" && $table'name eq "" )
		{
			$table'name = $name;
			&debug("dxftoobj: ignoring TABLE $table'name...");
		}

		next;
	};

	# handle all the other various group codes by setting some global
	# variables used throughout the dxf_ code

	$group == 1 && do { $text = $value; next; };

	$group == 3 && do { $other = $value; next };
	$group == 4 && do { $other2 = $value; next };

	$group == 5 && do { $handle = $value; next };

	$group == 6 && do { $linetype = $value; next };
	$group == 7 && do { $textstyle = $value; next };
	$group == 8 && do { $layer = $value; next };
	$group == 9 && do
	{
		# process the last variable "group"
		defined $variable && &dxf_variable;
		$variable = $value;
		next;
	};

	$group == 39 && do { $thickness = $value; next };

	$group == 49 && do { $repeated = $value; next };

	$group == 50 && do { $angle[0] = $value; next };
	$group == 51 && do { $angle[1] = $value; next };

	$group == 62 && do { $color = $value; next };

	$group == 66 && do { $follows = $value; next };

	# 10-18 - x[0] through x[8]
	$group > 9 && $group < 19 &&
		do{ $x[$group-10] = $value; next };

	# 20-28 - y[0] through y[8]
	$group > 19 && $group < 29 &&
		do{ $y[$group-20] = $value; next };

	# 30-38 - z[0] through z[8]
	$group > 29 && $group < 38 &&
		do{ $z[$group-30] = $value; next };

	$group == 38 && do
	{
		if ( ! $seen'elevation )
		{
			$seen'elevation = 1;
			&warning("DXF: obselete elevation code in use");
		}

		$elevation = $value;
		next;
	};

	# 40-48 - float[0] through float[8]
	$group > 39 && $group < 49 &&
		do { $float[$group-40] = $value; next };

	$group > 69 && $group < 80 &&
		do { $integer[$group-70] = int($value); next };

	$group == 210 && do { $extrude_x = $value; next };
	$group == 220 && do { $extrude_y = $value; next };
	$group == 230 && do { $extrude_z = $value; next };

	# Canvas 3.x use 999 to encode font names
	$group == 999 && do { $font = $value; next };

	&info("DXF: unsupported group (\"$group\"=\"$value\")");
}

# we shouldn't exit here

&warning("dxftoobj: EOF group not seen.  File possibly truncated.");
&dxf_eof;

exit 1;

sub
warning
{
	local($mesg) = @_;
	print STDERR "$mesg\n";
}

sub
info
{
	local($mesg) = @_;
	print STDERR "$mesg\n";
}

sub
debug
{
	local($mesg) = @_;
	$verbose && print STDERR "$mesg\n";
}

# DXF driver

# dxf_init -- prepare to output dxf entities
# BUGS
#   this does more than it should due to problems with BLOCK/INSERT
#
sub
dxf_init
{
	eval "&${driver}_open(\$output)";
	eval "&${driver}_remark(\"converted by dxftoobj (v \$dxftoobj_version)\")";
	eval "&${driver}_remark";
}

sub
dxf_variable
{
	$variable eq "\$EXTMIN" && do
	{
		$dxf_minext_x = $x[0];
		$dxf_minext_y = $y[0];
		$dxf_minext_z = $z[0];
	};

	$variable eq "\$EXTMAX" && do
	{
		$dxf_maxext_x = $x[0];
		$dxf_maxext_y = $y[0];
		$dxf_maxext_z = $z[0];
	};
}

sub
dxf_layer
{
	if ( $layer ne $lastlayer )
		{ eval "&${driver}_group($layer)"; }

	$lastlayer = $layer;
}

# dxf_circle -- draw an arc about the center $x[0], $y[0], $z[0] with radius
#              $float[0]
# BUGS
#   Doesn't understand the case where the arc is not on the XY plane
#
sub
dxf_circle
{
	++$dxf_ncircle;
	&dxf_layer;

	defined $thickness && &warning("CIRCLE: thickness unsupported");
	defined $elevation && &warning("CIRCLE: elevation unsupported");
	
	eval "&${driver}_remark(\"CIRCLE $x[0] $y[0] $z[0] $float[0]\")";
	eval "&${driver}_arc(\$x[0], \$y[0], \$z[0], \$float[0], 0.0, 360.0)";
}

# dxf_arc -- draw an arc about the center $x[0], $y[0], $z[0] with radius
#            $float[0], from $angle[0] to $angle[1].
# BUGS
#   Doesn't understand the case where the arc is not on the XY plane
#
sub
dxf_arc
{
	++$dxf_narc;
	&dxf_layer;

	defined $thickness && &warning("ARC: thickness unsupported");
	defined $elevation && &warning("ARC: elevation unsupported");

	eval "&${driver}_remark(\"ARC \$x[0] \$y[0] \$z[0] \$float[0] \$angle[0] \$angle[1]\")";
	eval "&${driver}_arc(\$x[0], \$y[0], \$z[0], \$float[0], \$angle[0], \$angle[1])";
}

# dxf_3dline -- draw a line from x[0], y[0], z[0] to x[1], y[1], z[1]
#
sub
dxf_3dline
{
	++$dxf_n3dline;
	&dxf_layer;

	defined $elevation && &warning("3DLINE: elevation unsupported");
	defined $elevation && &warning("3DLINE: thickness unsupported");

	eval "&${driver}_remark(\"3DLINE\")";
	eval "&${driver}_line(2, 0, \@x[0..1], \@y[0..1], \@z[0..1])";
}

# dxf_line -- draw a line from x[0], y[0], 0.0 to x[1], y[1], 0.0
#
sub
dxf_line
{
	++$dxf_nline;
	&dxf_layer;

	defined $elevation && &warning("LINE: elevation unsupported");

	if ( defined $thickness )
	{
		++$dxf_nline_thick;

		$x[2] = $x[1] + ($extrude_x * $thickness);
		$y[2] = $y[1] + ($extrude_y * $thickness);
		$z[2] = $z[1] + ($extrude_z * $thickness);

		$x[3] = $x[0] + ($extrude_x * $thickness);
		$y[3] = $y[0] + ($extrude_y * $thickness);
		$z[3] = $z[0] + ($extrude_z * $thickness);

		eval "&${driver}_remark(\"LINE (thick)\")";
		eval "&${driver}_face(4, \@x[0..3], \@y[0..3], \@z[0..3])";
	}
	else
	{
		eval "&${driver}_remark(\"LINE\")";
		eval "&${driver}_line(2, 0, \@x[0..1], \@y[0..1], 0.0, 0.0)";
	}
}

# dxf_trace -- draw a quadrilateral using x[0], y[0], z[0] through 
#              x[3], y[3], z[3]
sub
dxf_trace
{
	++$dxf_ntrace;
	&dxf_layer;

	defined $thickness && &warning("TRACE: thickness unsupported");

	defined $elevation && do
	{
		for( $i=0; $i<4; ++$i)
			{ $z[$i] += $elevation; }
	};

	eval "&${driver}_remark(\"TRACE\")";
	eval "&${driver}_face(4, \@x[0..3], \@y[0..3], \@z[0..3])";
}

# dxf_solid -- draw a quadrilateral or triangle using x[0], y[0], z[0]
#              through x[3], y[3], z[3]  (the last two points are the
#              same if its a triangle)
#
sub
dxf_solid
{
	++$dxf_nsolid;
	&dxf_layer;

	defined $thickness && &warning("SOLID: thickness unsupported");

	defined $elevation && do
	{
		for( $i=0; $i<4; ++$i)
			{ $z[$i] += $elevation; }
	};

	eval "&${driver}_remark(\"SOLID\")";

	if ( $x[2]==$x[3] && $y[2]==$y[3] && $z[2]==$z[3] )
	{
		# a triangle
		eval "&${driver}_face(3, \@x[0..2], \@y[0..2], \@z[0..2])";
	}
	else
	{
		# ordered as 2 triangles, fix order before sending to
		# driver_face
		eval "&${driver}_face(4,
			\$x[0], \$x[1], \$x[3], \$x[2],
			\$y[0], \$y[1], \$y[3], \$y[2],
			\$z[0], \$z[1], \$z[3], \$z[2])";
	}
}

# dxf_3dface -- draw a quadrilateral or triangle using x[0], y[0], z[0]
#               through x[3], y[3], z[3]  (the last two points are the
#               same if its a triangle)
# BUGS
#   seems very similar to solid, what is the difference
#
sub
dxf_3dface
{
	++$dxf_n3dface;
	&dxf_layer;

	eval "&${driver}_remark(\"3DFACE\")";
	if ( $x[2]==$x[3] && $y[2]==$y[3] && $z[2]==$z[3] )
		{eval "&${driver}_face(3, \@x[0..2], \@y[0..2], \@z[0..2])";}
	else
		{eval "&${driver}_face(4, \@x[0..3], \@y[0..3], \@z[0..3])";}
}

# dxf_point -- draw a point at x[0], y[0], z[0]
# BUGS
#   not implemented -- unsure how useful this is
#
sub
dxf_point
{
	++$dxf_npoint;

	defined $thickness && &warning("POINT: thickness unsupported");
	defined $elevation && &warning("POINT: elevation unsupported");

	&warning("POINT: ignoring point entity");
	#eval "&${driver}_point(\$x[0], \$y[0], \$z[0])";
}

# dxf_text -- place a string of text at location x[0], y[0], z[0]
# BUGS
#   no support for text attributes
#
sub
dxf_text
{
	return if $disable_text;
	&dxf_layer;

	++$dxf_ntext;
	eval "&${driver}_comm(\$text);";
	eval "&${driver}_text(\$text, \$x[0], \$y[0], \$z[0], \$float[0]);";
}


# dxf_flag -- check to see if $flag is set in $i
#
sub
dxf_flag
{
	local( $i, $flag ) = @_;
	return ( ($i & $flag) == $flag );
}

# dxf_polyline -- draw one of the following:
#     line           a list of vertices defining points along a path
#     polyface mesh  a list of vertices defining points and faces
#     polygon mesh   a list of vertices describing a rectangular mesh
#
# BUGS
#   meshes are always closed in M/N direction
#   smoothing points not supported
#   fitting points for mesh (group 75) are not supported
#   invisible edges on polygon mesh are ignored
#
sub
dxf_polyline
{
	++$dxf_npolyline;
	&dxf_layer;

	undef $polyline_thick;

	defined $thickness &&
		($polyline_thick = $thickness, ++$dxf_npolyline_thick);

	# flags for polyline group
	#   1 - closed (or closed in M direction for mesh)
	#   2 - has curve fitting points
	#   4 -
	#   8 - 3D polyline (if we make this polyline a face, this matters)
	#  16 - polygon mesh
	#  32 - closed in N direction
	#  64 - polyface mesh
	# 128 - linetype

	$polyline_type = "POLYLINE OPEN";
	$polyline_in = 1;
	$polyline_closed = 0;
	$polyline_threeD = 0;	# ignored
	$polyline_Mclosed = 0;	# closed in "m" direction (unsupported)
	$polyline_Nclosed = 0; 	# closed in "n" direction (unsupported)

	$v_n = 0;

	&dxf_flag($integer[0],1) && do
	{
		$integer[0] -= 1;

		$polyline_type = "POLYLINE CLOSED";
		$polyline_closed = 1;
		$polyline_Mclosed = 1;
	};

	&dxf_flag($integer[0],2) && do
	{
		$integer[0] -= 2;
		&warning("POLYLINE: curve fitting points ignored");
	};

	&dxf_flag($integer[0],8) && do
	{
		$integer[0] -= 8;
		$polyline_threeD = 1;
	};

	&dxf_flag($integer[0],32) && do
	{
		$integer[0] -= 32;

		$polyline_Nclosed = 1;
	};

	&dxf_flag($integer[0],128) && do
	{
		&warning("POLYLINE: ignoring linetype");
		$integer[0] -= 128;
	};

	# polyline is a polyface mesh

	&dxf_flag($integer[0],64) && do
	{
		$integer[0] -= 64;

		++$dxf_npolyline_fmesh;
		$polyline_type = sprintf("POLYFACE MESH (%dx%d)", $integer[2], $integer[1]); 
		$polyline_nvertices = $integer[1];
		$polyline_nfaces = $integer[2];
		$polyface'vertices = 1;
	};

	# polyline is a polygon mesh

	&dxf_flag($integer[0],16) && do
	{
		$integer[0] -= 16;

		++$dxf_npolyline_gmesh;

		$polyline_m = $integer[1];
		$polyline_n = $integer[2];
		$polyline_type = sprintf("POLYGON MESH (%dx%d)", $polyline_m, $polyline_n);
	};

	$integer[0] != 0 && do
	{
		&warning("POLYLINE:  unsupported bit codes in group 70 (flags=$integer[0], line=$__entity_line");
	};

	eval "&${driver}_remark(\"\$polyline_type\")";
}

# dxf_seqend -- mark the end of a vertex lsit of a polyline entity
#               or end the end of an attribute entry
#
sub
dxf_seqend
{
	# we only handle "plain" polyline here -- meshes are handled in
	# dxf_vertex

	# polyline_in will be one if we are processing a ponyline
	if ( $polyline_in == 1 )
	{
		# POLYLINE OPEN/CLOSED

		($polyline_type =~ /^POLYLINE/) && do
		{
			if ( defined $polyline_thick )
			{
				# XXX clean this up
				for( $i = 1; $i < $v_n; ++$i )
				{
					&obj_face( 4,
		$v_x[$i], $v_x[$i+1], $v_x[$i+1], $v_x[$i],
		$v_y[$i], $v_y[$i+1], $v_y[$i+1], $v_y[$i],
		$v_z[$i], $v_z[$i+1], $v_z[$i+1] + $polyline_thick, $v_z[$i] + $polyline_thick);
				}
			}
			else
			{
				eval "&${driver}_line(\$v_n,
					\$polyline_closed,
					\@v_x[1..\$v_n],
					\@v_y[1..\$v_n],
					\@v_z[1..\$v_n])";
			}
		};

		# POLYGON MESH

		($polyline_type =~ /^POLYGON MESH/) && do
		{
			for($i=1; $i<=$v_n; ++$i)
			{
				&obj_vertex($v_x[$i], $v_y[$i], $v_z[$i]);
			}

			for($i=($polyline_m*$polyline_n); $i>$polyline_n; $i-=$polyline_n)
			{

				for($j=0; $j<($polyline_n-1); ++$j)
				{
					printf OUT "f %d %d %d %d\n",
						(-$i)+$j,
						(-$i)+$j+$polyline_n,
						(-$i)+$j+1+$polyline_n,
						(-$i)+$j+1;
				}
			}
		};

		# POLYFACE MESH (handled in dxf_vertex)

		$polyline_in = 0;
	}
}

# dxf_vertex --
#
# 
sub
dxf_vertex
{
	++$dxf_nvertex;

	defined $elevation && ($z[0] += $elevation);
	#defined $thickness && &warning("VERTEX: thickness unsupported");

	# a point (or face) on a polyface mesh
	&dxf_flag($integer[0], 128) && do
	{
		if (&dxf_flag($integer[0], 64))
		{
			++$v_n;
			$v_x[$v_n] = $x[0];
			$v_y[$v_n] = $y[0];
			$v_z[$v_n] = $z[0];

			# this vertex actually is a vertex that will be
			# used later in defining a face on the mesh
			#
			&obj_vertex($x[0], $y[0], $z[0]);
			++$polyface'vertices; # save for relative refs
		}
		else
		{
			# defines a face on the polygon mesh
			# negative index values mean the edge is invisible
			# currently this is ignored
			#
			$integer[1] = &abs($integer[1]);
			$integer[2] = &abs($integer[2]);
			$integer[3] = &abs($integer[3]);
			print OUT "f ",
				$integer[1] - $polyface'vertices," ",
				$integer[2] - $polyface'vertices," ",
				$integer[3] - $polyface'vertices,"\n";
		}

		return;
	};

	# a point on a polygon mesh
	&dxf_flag($integer[0], 64) && do
	{
		++$v_n;
		$v_x[$v_n] = $x[0];
		$v_y[$v_n] = $y[0];
		$v_z[$v_n] = $z[0];

		return;
	};

	# a point on a 3D polyline
	&dxf_flag($integer[0], 32) && do
	{
		++$v_n;
		$v_x[$v_n] = $x[0];
		$v_y[$v_n] = $y[0];
		$v_z[$v_n] = $z[0];

		return;
	};

	# ???
	($integer[0] == 0) && do
	{
		++$v_n;
		$v_x[$v_n] = $x[0];
		$v_y[$v_n] = $y[0];
		$v_z[$v_n] = $z[0];

		return;
	};

	if ( $integer[0] != 0 )
		{ &info("VERTEX: unsupported flags $integer[0]"); }
}

# dxf_block -- defines a group of entities that may be INSERT'ed later
# BUGS
#   currently redefines output file to store groups as seperate files
#     that can later be read
#   doesn't save OUT (although this is possible but doesn't matter since
#     OUT isn't open when BLOCK's are defined)
#
sub
dxf_block
{
	++$dxf_nblock;
	&debug("BLOCK $name saved");

	eval "&${driver}_open(\$name)";
	eval "&${driver}_remark(\"BLOCK \$name\")";
}

# dxf_block -- ends a BLOCK
#
sub
dxf_endblk
{
	eval "&${driver}_remark(\"ENDBLK\")";
	eval "&${driver}_close";
}

# dxf_insert -- insert a BLOCK group, scaling by float[1], float[2], float[3]
#               and translating by x[0], y[0], and z[0]
# BUGS
#   BLOCK/ENDBLK sequence must occur before insert so group is available
#
sub
dxf_insert
{
	++$dxf_ninsert;
	&dxf_layer;

	&debug("INSERT: $name");
	eval "&${driver}_remark(\"INSERT $name\")";

	# if the scale for inserted entity is not specified, use 1,1,1
        !defined $float[1]  && ($float[1] = 1.0);
        !defined $float[2]  && ($float[2] = 1.0);
        !defined $float[3]  && ($float[3] = 1.0);

	eval "&${driver}_import(\$name, 
			\$x[0], \$y[0], \$z[0],
			\$float[1], \$float[2], \$float[3])" ||
		&warning("INSERT: failed to find \"$name\" for insertion");
}


# dxf_eof -- print statistics and close DXF file
#
sub
dxf_eof
{
	eval "&${driver}_close";

	&info("");
	&info("$dxf_npolyline POLYLINE entities");
	&info("\t$dxf_npolyline_thick POLYLINE (thick) entities");
	&info("\t$dxf_npolyline_gmesh POLYGON MESH entities");
	&info("\t$dxf_npolyline_fmesh POLYFACE MESH entities");
	&info("$dxf_nline LINE entities");
	&info("\t$dxf_nline_thick LINE entities (w/ thickness)");
	&info("$dxf_n3dline 3DLINE entities");
	&info("$dxf_n3dface 3DFACE entities");
	&info("$dxf_nvertex VERTEX entities");
	&info("$dxf_npoint POINT entities");
	&info("$dxf_ntext TEXT entities");
	&info("$dxf_nsolid SOLID entities");
	&info("$dxf_ncircle CIRCLE entities");
	&info("$dxf_narc ARC entities");
	&info("$dxf_ntrace TRACE entities");
	&info("$dxf_nblock BLOCK entities");
	&info("$dxf_ninsert INSERT entities");
}

sub
abs
{
	local($i) = @_;
	if ( $i < 0 ) { $i = -$i; }
	return $i;
}

#
# output driver routines
#
# Each driver should support:
#   driver_open     open driver (initialization, open file, etc)
#   driver_close    close driver (stats, close file)
#   driver_remark   emit a comment (if support by output format)
#   driver_line     generate a line
#   driver_face     generate a polygon
#   driver_arc      generate an arc
#   driver_import   import an external file
#   driver_group    change grouping information


# OBJ driver

sub
obj_open
{
	local($file) = @_;

	$obj_version = "0.9";

	$obj_vertices = 0;
	$obj_faces = 0;
	$obj_lines = 0;

	open(OUT, ">${file}.obj")
		|| die "obj: failed to open ${file}.obj: $!\n";

	&obj_remark(".obj driver (v $obj_version)");
	&obj_remark;
}

sub
obj_close
{
	&obj_remark("$obj_vertices points");
	&obj_remark("$obj_faces faces");
	&obj_remark("$obj_lines faces");

	close(OUT);
}

sub
obj_group
{
	local($group) = @_;
	print OUT "g $group\n";
}

sub
obj_vertex
{
	local($x, $y, $z) = @_;
	++$obj_vertices;
	printf OUT "v %f %f %f\n", $x, $y, $z;
}

sub
obj_remark
{
	local( $remark ) = @_;
	printf OUT "# $remark\n";
}

sub
obj_line
{
	local( $n, $closed, @pt) = @_;

	local($i);
	++$obj_lines;

	for( $i = 0; $i < $n; ++$i )
	{
		&obj_vertex($pt[$i], $pt[$i+($n)], $pt[$i+($n*2)]);
	}

	# the .obj format doesn't support lines with more than 500 points
	# so long lines should be broken into equal lengths

	# XXX fix w/ recursion?
	#   obj_line($n/2, $open, @pt[0..-n/2], @pt[n..(n+n/2)], ..
	#   obj_line($n/2, $open, @pt[0..-n/2], @pt[n..(n+n/2)], ..

	if ( $n > 500 )
	{
printf STDERR "line w/ more than 500 points (%d)\n", $n;
		print OUT "l";
		for( $i = $n; $i > 500; --$i )
		{
			printf OUT " %d", -$i;
		}
		printf OUT "\n";

		$n = 500;
	}

	print OUT "l";
	for( $i = $n; $i > 0; --$i )
	{
		printf OUT " %d", -$i;
	}

	$closed && printf OUT " %d", -$n;
	printf OUT "\n";
}

sub
obj_face
{
	local( $n, @pt) = @_;

	local($i);
	++$obj_faces;

	for( $i = 0; $i < $n; ++$i )
		{ &obj_vertex($pt[$i], $pt[$i+$n], $pt[$i+($n*2)]); }

	print OUT "f";
	for( $i = $n; $i > 0; --$i )
		{ printf OUT " %d", -$i; }

	print OUT "\n";
}

sub
obj_text
{
	local($text, $x, $y, $z, $scale) = @_;

	# calls Wavefront's bit program to generate a text string
	# the file is then "imported" to the current output file
	# applying the needed translation and scales
	# this generates large/complex objects due to the nature of
	# fonts
	&debug("OBJ: calling bit to build \"$text\" object");

	if ( !defined $ENV{'WF_AV_DIR'} )
	{
		&warning("obj_text: WF_AV_DIR not set; cannot convert text");
		return 0;
	}

	$WF_AV_DIR=$ENV{'WF_AV_DIR'};
		
	unlink("text.obj");
	system("echo \"$text\" | $WF_AV_DIR/bin/bit -L -a 10.0 -o text -q 0 -s swi_ult > /dev/null");

	&obj_import("text", $x, $y, $z, $scale/3, $scale/3, $scale/3);
}

sub
obj_import
{
	local( $name, $tx, $ty, $tz, $sx, $sy, $sz ) = @_;

	$n = 0;

	open(IMPORT, "<$name.obj") || return 0;
	while(<IMPORT>)
	{
		chop;
		/^v/ && do
		{
			($dummy, $x, $y, $z, $crap) = split;
			&obj_vertex($tx+($x*$sx), $ty+($y*$sy), $tz+($z*$sz));
			++$n;
		};

		# pass comments and group information
		/^[g#]/ && printf OUT "%s\n", $_;

		/^[lf]/ && do
		{
                        ($element, @indices) = split;
                        printf OUT "%s", $element;
                        foreach $index ( @indices )
                        {
				if ($index < 0)
					{ printf OUT " %d", $index; }
				else
					{ printf OUT " %d", ($index-$n)-1; }
                        }
                        print OUT "\n";
		};
	}
	close(IMPORT);
	return 1;
}

# obj_arc -- draw an arc.  arc_segments controls resolution of arc segments
#
# BUGS
#   Converts arcs to a series of short line-segments
#   Arc length (not $arc_segments) should control arc resolution
#
sub
obj_arc
{
	local($x, $y, $z, $radius, $from, $to) = @_;

	if ( $to < $from ) { $to += 360.0; }

	# convert to radians...
	$degrees_to_radians = 3.1415/180.0;
	$from = $from * $degrees_to_radians;
	$to = $to * $degrees_to_radians;

	$delta = ($to - $from)/($arc_segments-1);

	for( $i = 0; $i < $arc_segments; ++$i )
	{
		$theta = $from + ($i * $delta);
		&obj_vertex($x+($radius*cos($theta)), $y+($radius*sin($theta)), $z);
	}

	print OUT "l ";
	for( $i = $arc_segments; $i > 0; --$i )
	{
		printf OUT "%d ", -$i;
	}
	print OUT "\n";
}

# PS driver routines

sub
ps_open
{
	local($file) = @_;

	$ps_version = "0.5";

	open(OUT, ">${file}.ps")
		|| die "ps: failed to open ${file}.ps: $!\n";

	print OUT <<EOF;
%!

% PostScript driver version $ps_version

/centre % x y message
{
	/message    exch def
	/y          exch def
	/x          exch def

	x message stringwidth pop 2 div add
	y
	moveto
	message show
} def

0 setlinewidth
0 setgray

/Textfont
    /Helvetica-Bold findfont def

EOF

	# a portrait page is 612x792 points

	$dx = &abs($dxf_maxext_x - $dxf_minext_x);
	$dy = &abs($dxf_maxext_y - $dxf_minext_y);
	# $dz = &abs($dxf_maxext_z - $dxf_minext_z);

	$sx = 612.0/$dx;
	$sy = 792.0/$dy;
	
	if ($sx < $sy)
		{ $s = $sx * 0.90; }
	else
		{ $s = $sy * 0.90; }

	$s = 10.0;

	$cx = ($dxf_maxext_x - $dxf_minext_x)/2;
	$cy = ($dxf_maxext_y - $dxf_minext_y)/2;
	# $cz = ($dxf_maxext_z - $dxf_minext_z)/2;

	printf OUT "%f %f scale\n", $s, $s;
	printf OUT "%f %f translate\n", (306+$cx)*$s, (396-$cy)*$s;
}


sub
ps_close
{
	printf OUT "showpage\n";
	close(OUT);
}

sub
ps_group
{
	# no group support in postscript
}

sub
ps_remark
{
	local( $remark ) = @_;
	printf OUT "%% %s\n", $remark;
}

sub
ps_moveto
{
	local($x, $y) = @_;
	printf OUT "%f %f moveto\n", $x, $y;
}

sub
ps_lineto
{
	local($x, $y) = @_;
	printf OUT "%f %f lineto\n", $x, $y;
}

sub
ps_line
{
	local( $n, $closed, @pt) = @_;
	
	&ps_moveto($pt[0], $pt[0+$n]);
	for($i=1; $i<n; ++$n)
		{ &ps_lineto($pt[$i], $pt[$i+$n]); }
	if ($closed)
		{ &ps_moveto($pt[0], $pt[$n]); }
	printf OUT "stroke\n";
}

sub
ps_face
{
	local($n, @pt) = @_;

	&ps_moveto($pt[0], $pt[0+$n]);

	for( $i=1; $i<$n; ++$i)
		{ &ps_lineto($pt[$i], $pt[$i+$n]); }

	&ps_lineto($pt[0], $pt[$n]);
	printf OUT "stroke\n";
}

sub
ps_text
{
	local($text, $x, $y, $z, $scale) = @_;
	
	printf OUT <<EOF;
Textfont $scale scalefont setfont
$x $y ($text) centre
EOF
	return;
}

sub
ps_arc
{
	local($x, $y, $z, $radius, $from, $to) = @_;

	if ( $to < $from ) { $to += 360.0; }

	# convert to radians...
	$degrees_to_radians = 3.1415/180.0;
	$from = $from * $degrees_to_radians;
	$to = $to * $degrees_to_radians;

	$delta = ($to - $from)/($arc_segments-1);

	for( $i = 0; $i < $arc_segments; ++$i )
	{
		$theta = $from + ($i * $delta);
		if ( $i )
			{ printf OUT "%f %f lineto\n",
				$x+($radius*cos($theta)),
				$y+($radius*sin($theta)); }
		else
			{ printf OUT "%f %f moveto\n",
				$x+($radius*cos($theta)),
				$y+($radius*sin($theta)); }
	}

	printf OUT "stroke\n";
}

sub
ps_import
{
	local( $name, $tx, $ty, $tz, $sx, $sy, $sz ) = @_;

	open(IMPORT, "<$name.ps") || return 0;
	while(<IMPORT>)
	{
		chop;
		/lineto/ && do
		{
			($x, $y, $type) = split;
			&ps_lineto($tx+($x*$sx), $ty+($y*$sy));
		};

		/moveto/ && do
		{
			($x, $y, $type) = split;
			&ps_moveto(tx+($x*$sx), $ty+($y*$sy));
		};

		/stroke/ && printf OUT "%s\n", $_;
	}
	close(IMPORT);
	return 1;
}

# MDL driver routines

sub
mdl_open
{
	local($file) = @_;

	$mdl_version = "0.1";
	$mdl_name = "default";

	open(OUT, ">${file}.mdl")
		|| die "mdl: failed to open ${file}.mdl: $!\n";

	&mdl_remark(".mdl driver (v $mdl_version)");
}

sub
mdl_group
{
	local($group) = @_;

	$mdl_name = $group;
}

sub
mdl_close
{
	close(OUT);
}

sub
mdl_vertex
{
	local($x, $y, $z) = @_;
	printf OUT " vertex %f %f %f\n", $x, $y, $z;
}

sub
mdl_remark
{
}

sub
mdl_line
{
	local( $n, $closed, @pt) = @_;
	
	printf OUT "object line\n";
	printf OUT " name %s\n", $mdl_name;
	printf OUT " doublesided\n";

	for($i=0; $i<n; ++$i)
		{ &mdl_vertex($pt[$i], $pt[$i+$n], $pt[$i+($n*2)]); }
	if ($closed)
		{ &mdl_vertex($pt[0], $pt[$n], $pt[$n*2]); }
}

sub
mdl_face
{
	local($n, @pt) = @_;

	printf OUT "object face\n";
	printf OUT " name %s\n", $mdl_name;
	printf OUT " doublesided\n";

	for($i=0; $i<$n; ++$i)
		{ &mdl_vertex($pt[$i], $pt[$i+$n], $pt[$i+($n*2)]); }
}

sub
mdl_text
{
}

sub
mdl_import
{
	local( $name, $tx, $ty, $tz, $sx, $sy, $sz ) = @_;

	open(IMPORT, "<$name.mdl") || return 0;
	while(<IMPORT>)
	{
		chop;
		/^ vertex/ && do
		{
			($dummy, $dummy, $x, $y, $z) = split(/\s+/);
			&mdl_vertex($tx+($x*$sx), $ty+($y*$sy), $tz+($z*$sz));
			next;
		};

		printf OUT "%s\n", $_;
	}
	close(IMPORT);
	return 1;
}

sub
mdl_arc
{
	return;

	local($x, $y, $z, $radius, $from, $to) = @_;

	if ( $to < $from ) { $to += 360.0; }

	# convert to radians...
	$degrees_to_radians = 3.1415/180.0;
	$from = $from * $degrees_to_radians;
	$to = $to * $degrees_to_radians;

	$delta = ($to - $from)/($arc_segments-1);

	for( $i = 0; $i < $arc_segments; ++$i )
	{
		$theta = $from + ($i * $delta);
		&obj_vertex($x+($radius*cos($theta)), $y+($radius*sin($theta)), $z);
	}

	print OUT "l ";
	for( $i = $arc_segments; $i > 0; --$i )
	{
		printf OUT "%d ", -$i;
	}
	print OUT "\n";
}
